@@ -21,7 +21,8 @@
 #ifndef LLD_ELF_LTO_H
 #define LLD_ELF_LTO_H
 
-#include "lld/Core/LLVM.h"
+#include "lld/Common/LLVM.h"
+#include "llvm/ADT/DenseSet.h"
 #include "llvm/ADT/SmallString.h"
 #include <memory>
 #include <vector>
@@ -50,6 +51,7 @@ private:
   std::unique_ptr<llvm::lto::LTO> LTOObj;
   std::vector<SmallString<0>> Buff;
   std::vector<std::unique_ptr<MemoryBuffer>> Files;
+  llvm::DenseSet<StringRef> UsedStartStop;
 };
 } // namespace elf
 } // namespace lld
